import data.nat.prime
import data.real.basic
import data.padics.padic_norm
import ring_theory.principal_ideal_domain
import data.real.cau_seq
import analysis.complex.exponential
import tactic.apply



section


open_locale classical
open list nnreal option
-- TODO i thought this worked before with only ordered_semiring
lemma prod_lt_one_lt_one {α : Type*} [linear_ordered_semiring α] [canonically_ordered_monoid α] {l : list α} (hl : l ≠ []) (hlt : ∀ (a : α) (ha : a ∈ l), a < 1) : l.prod < 1 :=
begin
   induction l,
   simpa only [prod_nil, not_true, ne.def, prod_nil, ne.def] using hl,
   rw prod_cons,
   have hd_lt_one := hlt l_hd (by simp),
   by_cases l_tl = [],
   simp only [h, hd_lt_one, prod_nil, mul_one],
   have := l_ih h (begin intros b hb, convert hlt b _, rw mem_cons_iff, simp only [hb, or_true], end),
   library_search,
   convert mul_lt_mul'' hd_lt_one this (zero_le l_hd) (zero_le _),
   rw mul_one,
end

end

open padic_norm nat
open_locale real

def equivalent (abv1 abv2 : ℚ → ℚ) [is_absolute_value abv1] [is_absolute_value abv2] := ∃ (r : ℝ) (hr : 0 < r), (coe ∘ abv1 : ℚ → ℝ) = ((λ x : ℚ, x ^ r) ∘ abv2)

instance is_absolute_value.setoid : setoid ({f : ℚ → ℚ // is_absolute_value f}) :=
{ r := λ x1 x2, @equivalent x1.1 x2.1 x1.2 x2.2,
  iseqv := begin
  split,
  {
    dunfold reflexive,
    intros,
    use ⟨1, by norm_num⟩,
  },
  split,
  {
    dunfold symmetric,
    intros,
    obtain ⟨r, hr, _⟩ := a,
    use [r⁻¹, inv_pos hr],
    ext,
    simp,

    have := congr_fun a_h_h x_1,
    simp at this,
    rw this,
    rw ← real.rpow_mul,
    rw mul_inv_cancel (ne_of_gt hr),
    simp,
    norm_cast,
    haveI := y.2,
    exact is_absolute_value.abv_nonneg y.val _,
  },
  {
    dunfold transitive,
    intros,
    obtain ⟨r, hr, _⟩ := a,
    obtain ⟨s, hs, _⟩ := a_1,
    use [s * r, mul_pos hs hr],
    ext,
    simp,

    have := congr_fun a_h_h x_1,
    simp at this,
    rw this,

    have := congr_fun a_1_h_h x_1,
    simp at this,
    rw this,

    rw ← real.rpow_mul,
    norm_cast,
    haveI := z.2,
    exact is_absolute_value.abv_nonneg z.val _,
  }
  end }

variables {α : Type*} [linear_ordered_field α] [archimedean α]
lemma pow_unbounded_below_of_lt_one (x : α) {y : α}
    (hy : 0 < y) (hx : 0 < x) (hy1 : y < 1) : ∃ n : ℕ, y ^ n < x :=
begin
  obtain n := pow_unbounded_of_one_lt (1/x) (one_lt_one_div hy hy1 : 1 < 1/y),
  use n,
  rw div_pow at h,
  simp [-one_div_eq_inv] at h,
  exact lt_of_one_div_lt_one_div hx h,
  exact ne_of_gt hy,
end

def trivial_abv : ℚ → ℚ := λ x, if x = 0 then 0 else 1
#check rat.mk
lemma rat_mk_eq_div (x_num : ℤ) (x_denom : ℕ) (x_pos : 0 < x_denom) (x_cop : coprime (int.nat_abs x_num) x_denom)
: ({num := x_num, denom := x_denom, pos := x_pos, cop := x_cop}:ℚ) = (↑ x_num)/(↑ x_denom) := begin
norm_cast,
rw rat.mk,
rw rat.mk_nat,
rw dif_neg,
rw rat.mk_pnat,
{
  simp [x_cop.gcd_eq_one],
},
exact ne_of_gt x_pos,
end

instance : is_absolute_value trivial_abv :=
{ abv_nonneg := λ x, begin rw [trivial_abv], dsimp, split_ifs, exact le_refl 0, exact zero_le_one, end,
  abv_eq_zero := λ x, begin rw [trivial_abv], dsimp, split, contrapose!, intro, rw if_neg a, norm_num, intro, rw if_pos a, end,
  abv_add := λ x y, begin rw [trivial_abv], dsimp, split_ifs, any_goals { linarith,}, exfalso, rw [h_1,h_2] at h, simpa using h, end,
  abv_mul := λ x y, begin rw [trivial_abv], dsimp, split_ifs, any_goals {norm_num}, finish, finish, finish, finish, end }

def is_equivalent := is_absolute_value.setoid.r

theorem rat_abs_val_eq_of_eq_on_pnat (abv abv2 : ℚ → ℚ) [habv : is_absolute_value abv] [habv2 : is_absolute_value abv2] (h : ∀ n : ℕ+, abv n = abv2 n)
 : abv = abv2 :=
begin
  ext,
  suffices : ∀ z : ℤ, abv z = abv2 z,
  begin
    induction x,
    rw rat_mk_eq_div,
    rw is_absolute_value.abv_div abv,
    rw is_absolute_value.abv_div abv2,
    rw this x_num,
    have := this (↑ x_denom),
    norm_cast at this,
    rw this,
  end,
  intro x,
  suffices : ∀ z : ℕ, abv z = abv2 z,
  begin
    induction x,
    simp [this],
    push_cast,
    rw is_absolute_value.abv_neg abv,
    rw is_absolute_value.abv_neg abv2,
    norm_cast,
    exact this _,
  end,
  intro x,
  cases x,
  simp,
  rw is_absolute_value.abv_zero abv,
  rw is_absolute_value.abv_zero abv2,
  have := h ⟨succ x_1, succ_pos _⟩,
  simp [-add_comm] at this,
  simp [-add_comm],
  rw this,
end


example (n : ℕ) (a_1 : (↑ (n) :ℝ) = 0) :
n = 0 :=
begin
exact nat.cast_eq_zero.mp a_1, -- TODO did lib search find?
end


lemma rat_abs_val_bdd_padic (abv : ℚ → ℚ) [habv : is_absolute_value abv]
  (hnontriv : ¬is_equivalent ⟨abv, habv⟩ ⟨trivial_abv, trivial_abv.is_absolute_value⟩)
  (int_bdd : ∃ (B : ℚ), ∀ (z : ℕ), abv ↑z ≤ B) :
   ∃ (p) [hp : nat.prime p],
    is_equivalent ⟨abv, habv⟩ ⟨padic_norm p, by haveI := rat.discrete_linear_ordered_field;
    apply_instance⟩ :=
begin
    obtain ⟨B, int_lt_B⟩ := int_bdd,
    have all_nat_le_one : ∀ z : ℕ, abv z ≤ 1 :=
    begin
      contrapose! int_lt_B,
      obtain z := int_lt_B,
      obtain n := pow_unbounded_of_one_lt B int_lt_B_h,
      use z ^ n,
      push_cast,
      rwa is_absolute_value.abv_pow abv,
    end,
    have all_int_le_one : ∀ z : ℤ, abv z ≤ 1 :=
    begin
      intro,
      induction z,
      exact all_nat_le_one _,
      push_cast,
      rw is_absolute_value.abv_neg abv,
      exact_mod_cast all_nat_le_one _,
    end,
    have : ∃ z : ℕ+, abv z < 1 :=
    begin
      by_contra,
      push_neg at a,
      have : ∀ z : ℕ+, abv z = 1 := λ z, le_antisymm (all_nat_le_one z) (a z),
      apply hnontriv,
      use 1,
      use zero_lt_one,
      ext,
      simp,
      suffices : abv = trivial_abv,
      begin
        exact congr_fun this x,
      end,
      apply rat_abs_val_eq_of_eq_on_pnat,
      intro,
      unfold trivial_abv,
      rw if_neg,
      exact this n,
      have := ne_of_gt n.2,
      intro,
      simpa using a_1,
    end,
    obtain ⟨z, hz⟩ := this,
    have : ∃ p ∈ factors z, abv p < 1 :=
    begin
      by_contra,
      push_neg at a,
      have : (factors (z.val)).prod = z.val := prod_factors z.2,
      have := congr_arg (abv ∘ coe) this,
      simp at this,
      rw (sorry : ↑(list.prod (factors (z.val))) = list.prod ((factors z.val).map (coe : ℕ → ℚ))) at this,
      rw is_absolute_value.abv_prod abv at this,
      sorry,
    end,
    obtain ⟨p, p_fact, abv_p_lt_one⟩ := this,
    have : ∀ q (hpq : coprime p q), abv q = 1 :=
    begin
      by_contra,
      push_neg at a,
      obtain q := a,
      have : ∃ e : ℕ, abv (p ^ e) < 1/2 ∧ abv (q^e) < 1/2 :=
      begin
       conv
       {congr,funext, rw is_absolute_value.abv_pow abv,rw is_absolute_value.abv_pow abv,},
        have : abv q < 1 := lt_of_le_of_ne (all_int_le_one q) a_h.2,
        obtain n1 := pow_unbounded_below_of_lt_one (1/2) _ _ this,
        obtain n2 := pow_unbounded_below_of_lt_one (1/2) _ _ abv_p_lt_one,
        use max n1 n2,
        sorry,
        sorry,
        sorry,
        sorry,
        sorry,
      end,
      obtain ⟨e, pe_half, qe_half⟩ := this,
      have := gcd_eq_gcd_ab (p^e) (q^e),
      have gcdpq_eq_one : nat.gcd (p^e) (q^e) = 1 := sorry,
      rw gcdpq_eq_one at this,
      have one_eq_bezout := congr_arg (abv ∘ coe) this,
      simp at one_eq_bezout,
      rw is_absolute_value.abv_one abv at one_eq_bezout,
      have := is_absolute_value.abv_add abv (↑p ^ e * ↑(gcd_a (p ^ e) (q ^ e)))
        (↑q ^ e * ↑(gcd_b (p ^ e) (q ^ e))),
      simp at this,
      rw ← one_eq_bezout at this,
      rw is_absolute_value.abv_mul abv at this,
      rw is_absolute_value.abv_mul abv at this,
      have gcd_a_lt := all_int_le_one (gcd_a (p ^ e) (q ^ e)),
      have gcd_b_lt := all_int_le_one (gcd_b (p ^ e) (q ^ e)),
      have : abv (↑p ^ e) * abv ↑(gcd_a (p ^ e) (q ^ e)) < 1 / 2 :=
        (mul_one (1/2)) ▸ (mul_lt_mul pe_half gcd_a_lt ((is_absolute_value.abv_pos abv).mpr _) (by norm_num)),
      have : abv (↑q ^ e) * abv ↑(gcd_b (p ^ e) (q ^ e)) < 1 / 2 :=
        (mul_one (1/2)) ▸ (mul_lt_mul qe_half gcd_b_lt ((is_absolute_value.abv_pos abv).mpr _) (by norm_num)),
      linarith,
      norm_cast,
      sorry,
      sorry,
    end,
    have hp : p.prime := mem_factors p_fact,
    use [p, hp],
    let α := abv p,
    use [-(real.log α / real.log p)],
    have : 0 < -(real.log ↑α / real.log ↑p) :=
    begin
      have log_alpha_neg : real.log α < 0 := real.log_neg _ (by exact_mod_cast abv_p_lt_one),
      work_on_goal 1 {
        norm_cast,
        exact (is_absolute_value.abv_pos abv).mpr (by exact_mod_cast nat.prime.ne_zero hp),
      },
      have log_p_pos : 0 < real.log p := real.log_pos (by exact_mod_cast hp.one_lt),
      rw neg_pos,
      exact div_neg_of_neg_of_pos log_alpha_neg log_p_pos,
    end,
    use this,
    ext,
    simp,
    suffices : ∀ x : ℤ, (abv ↑x : ℝ) = ((↑(padic_norm p x)) ^ -((real.log ↑α) / (real.log ↑p))),
    begin
      induction x,
      have hn := this x_num,
      have hd := this x_denom,
      have := congr_arg2 ((/) : ℝ → ℝ → ℝ) hn hd,
      norm_cast at this,
      rw ← is_absolute_value.abv_div abv at this,
      conv_rhs at this { rw div_eq_mul_inv,},
      simp at this,
      rw real.rpow_inv at this,
      conv_rhs at this { congr, skip,
      rw ← real.inv_rpow sorry,},
      rw ← real.mul_rpow at this,
      conv_rhs at this { rw ← div_eq_mul_inv,},
      norm_cast at this,
      rw ← @padic_norm.div p hp at this,
      convert this,
      convert rat_mk_eq_div x_num x_denom _ _,
      norm_cast,
      convert rat_mk_eq_div x_num x_denom _ _, -- LOL
      any_goals {norm_cast, try {rw inv_nonneg}, exact @padic_norm.nonneg p _,},
    end,
    intro,
    haveI : principal_ideal_domain ℤ := by apply_instance,
    haveI : unique_factorization_domain ℤ := principal_ideal_domain.to_unique_factorization_domain,

    apply unique_factorization_domain.induction_on_prime x_1,
    {
      simp [nat.cast_zero],
      unfold_coes,
      rw ← (by norm_cast : (0 : ℚ) = (int.cast (0 : ℤ))),
      rw padic_norm.zero,
      rw is_absolute_value.zero abv,
    },
    {
      intros,
      sorry
    },
    {
      intros,
      by_cases hpp1 : p = p_1,
    }

end

.

open finset list
def base_expand (b : ℕ) (l : list ℕ) : ℕ := (fintype.elems (fin l.length)).sum (λ i : fin (l.length), (l.nth_le i i.is_lt) * (b ^ (i:ℕ)))

@[simp]
lemma base_exp_nil {b : ℕ} : base_expand b [] = 0 := rfl
@[simp]
lemma base_exp_cons {b a : ℕ} (l : list ℕ) : base_expand b (a :: l) = a + b * base_expand b l :=
begin
  unfold base_expand,sorry;
  conv_lhs {congr, simp [length_cons a l],},
end

def base_expansion (b : ℕ) (hb : 1 < b) : ℕ → list ℕ
| 0       := []
| (n + 1) := have (n + 1) / b < n + 1, from div_lt_self (succ_pos n) hb,
               ((n + 1) % b) :: (base_expansion ((n+1)/b))
#eval base_expansion 2 (by norm_num) 4

@[simp]
lemma base_expansion_zero (b : ℕ) (hb : 1 < b) : base_expansion b hb 0 = [] := rfl

lemma base_expansion_spec (b : ℕ) (hb : 1 < b) : ∀ (n : ℕ), base_expand b (base_expansion b hb n) = n
| 0       := by simp
| (n + 1) := have (n + 1) / b < n + 1, from div_lt_self (succ_pos n) hb,
            begin
            simp [base_expansion],
            rw base_expansion_spec,
              exact nat.mod_add_div _ _,
             end

lemma base_expansion_le (b : ℕ) (hb : 1 < b) : ∀ (n : ℕ) (x) (h : x ∈ (base_expansion b hb n)), x < b
| 0       := by simp
| (n + 1) := have (n + 1) / b < n + 1, from div_lt_self (succ_pos n) hb,
            begin
            simp [base_expansion],
            split,
            exact nat.mod_lt (n + 1) (lt_of_succ_lt hb),
            exact base_expansion_le _,
             end

open real
lemma base_expansion_length (b : ℕ) (hb : 1 < b) : ∀ (n : ℕ), ((base_expansion b hb n).length : ℝ) ≤ log n / log b + 1
| 0       := by simp; exact zero_le_one
| (n + 1) := have (n + 1) / b < n + 1, from div_lt_self (succ_pos n) hb,
            begin
            simp [base_expansion, -add_comm],
            convert base_expansion_length ((n + 1) / b),
            suffices : log (↑n + 1) = log ↑((n + 1) / b) + log ↑b,
            by sorry,
            rw (by sorry :  (↑((n + 1) / b) :ℝ) = ↑(n+1)/↑b),
            rw log_div_eq_log_sub_log,
            simp only [cast_one, cast_add, add_comm, add_neg_cancel_left, sub_eq_add_neg],
            all_goals{norm_cast, linarith,},
             end

lemma ne_zero_one_one_lt (n0 : ℕ)
  (n0_ne_0 : n0 ≠ 0) (n0_ne_1 : n0 ≠ 1) : 1 < n0 := by omega

example (m n : ℕ) (h : m ≤ n): (↑( n - m) : ℝ) = (↑ n :ℝ) - (↑ m:ℝ) :=
begin
library_search
end

lemma coe_sum {α : Type*} {fs : finset α} {f : α → ℕ} : (↑ (sum fs f) : ℚ) = sum fs (coe ∘ f) :=
begin
sorry,
end

lemma coe_sum2 {α : Type*} {fs : finset α} {f : α → ℚ} : (↑ (sum fs f) : ℝ) = sum fs (coe ∘ f) :=
begin
sorry,
end
lemma rat_abs_val_unbdd_real (abv : ℚ → ℚ) [habv : is_absolute_value abv]
  --(hnontriv : ¬is_equivalent ⟨abv, habv⟩ ⟨trivial_abv, trivial_abv.is_absolute_value⟩)
  (int_bdd : ¬∃ (B : ℚ), ∀ (z : ℕ), abv ↑z ≤ B) :
  is_equivalent ⟨abv, habv⟩ ⟨abs, by apply_instance⟩ :=
begin
  push_neg at int_bdd,
  specialize int_bdd 1,
  have n0_spec := nat.find_spec int_bdd,
  set n0 := nat.find int_bdd,
  have n0_ne_0 : n0 ≠ 0 := by by_contra; simp at a; simp [a, is_absolute_value.abv_zero abv] at n0_spec; linarith,
  have n0_ne_1 : n0 ≠ 1 := by by_contra; simp at a; simp [a, is_absolute_value.abv_one abv] at n0_spec; linarith,
  have one_lt_n0 : 1 < n0 := ne_zero_one_one_lt _ n0_ne_0 n0_ne_1,
  have : ∀ n : ℕ, (↑ (abv n) : ℝ) ≤ ↑n ^ (real.log (abv n0) / real.log n0) :=
  begin
    intro,
    have h := base_expansion_spec n0 one_lt_n0 n,
    unfold base_expand at h,
    set l := base_expansion n0 one_lt_n0 n,
    replace h := congr_arg (abv ∘ coe) h,
    simp at h,
    rw coe_sum at h,
    have := is_absolute_value.abv_finset_sum_le abv (fintype.elems (fin (length l))) (coe ∘ λ (i : fin (length l)), ((nth_le l ↑i i.is_lt) * n0 ^ i.val)),
    /-have : (↑(abv (sum (fintype.elems (fin (length l))) (coe ∘ λ (i : fin (length l)), nth_le l ↑i _ * n0 ^ i.val))) : ℝ) ≤
      ↑(sum (fintype.elems (fin (length l))) (abv ∘ coe ∘ λ (i : fin (length l)), nth_le l ↑i _ * n0 ^ i.val)) := begin
      norm_cast,
      exact this,
    end,-/
    erw h at this,
    simp [is_absolute_value.abv_mul abv, is_absolute_value.abv_pow abv] at this,
    have le_aux : ∀ (x : fin l.length) (hx : x ∈ (fintype.elems (fin (length l)))), (
      (λ (x : fin (length l)), abv ↑(nth_le l ↑x x.is_lt) * abv ↑n0 ^ x.val) x ≤
      (λ (x : fin (length l)), abv ↑n0 ^ x.val) x) :=
    begin
      intros,
      simp,
      have : nth_le l ↑x x.is_lt ∈ l := nth_le_mem l ↑x _,
      have : nth_le l ↑x x.is_lt < n0 := base_expansion_le n0 one_lt_n0 n (nth_le l ↑x (x.is_lt)) this,
      have n0_min := nat.find_min int_bdd,
      specialize n0_min this,
      simp at n0_min,
      have : 0 ≤ abv ↑n0 ^ x.val := pow_nonneg (is_absolute_value.abv_nonneg _ _) _,

      exact mul_le_of_le_one_left this n0_min,
    end,
    have := le_trans this (finset.sum_le_sum le_aux),
    have le_aux : ∀ (x : fin l.length) (hx : x ∈ (fintype.elems (fin (length l)))), (
      (λ (x : fin (length l)), (abv ↑n0 : ℝ) ^ x.val) x ≤
      (λ (x : fin (length l)), (abv ↑n0 : ℝ) ^ (log ↑n / log ↑n0)) x) :=
    begin
      intros,
      simp,
      norm_cast,
      have := rpow_le_rpow_of_exponent_le (by exact_mod_cast le_of_lt n0_spec : (1 : ℝ) ≤ abv ↑n0)
      (begin
      transitivity (↑(l.length - 1) : ℝ),
      have := (le_pred_of_lt (x.is_lt)),
      exact_mod_cast this,
      have : 1 ≤ l.length := begin apply le_of_not_lt, intro, have := x.is_lt, have : (x.val < 0) := by linarith, exact nat.not_lt_zero _ this,  end,

      rw (cast_sub this : (↑(l.length - 1) : ℝ) = l.length - ↑ 1),
      rw cast_one,
      convert sub_le_sub_right (base_expansion_length n0 one_lt_n0 n) 1,
      symmetry,
      exact add_sub_cancel _ _,
      end : (x.val : ℝ) ≤ log ↑n / log ↑n0),
      convert this,
      push_cast,
      sorry,
    end,
    rw [← @rat.cast_le ℝ _ _] at this,
    push_cast at this,
    rw coe_sum2 at this,
    simp at this,
    push_cast at le_aux,
    have := le_trans (by exact_mod_cast this) (finset.sum_le_sum le_aux),
    rw finset.sum_const at this,
    rw (rfl : fintype.elems (fin l.length) = ⟨list.fin_range l.length, list.nodup_fin_range _⟩) at this,
    have card_is := fintype.card_fin l.length,
    rw fintype.card at card_is,
    rw univ at card_is,
    rw card_is at this,
    simp [l] at this,
    have := le_trans this base_expansion_length n0 one_lt_n0 n,
  end
end
#check fintype.elems (fin 5).


.

/- Ostrowski -/
theorem rat_abs_val_p_adic_or_real (abv : ℚ → ℚ) [habv : is_absolute_value abv]
(hnontriv : ¬ (is_equivalent ⟨abv, habv⟩ ⟨trivial_abv, by apply_instance⟩))
: is_equivalent ⟨abv, habv⟩ ⟨abs, by apply_instance⟩
∨ ∃ (p) [hp : nat.prime p], is_equivalent ⟨abv, habv⟩ ⟨padic_norm p, by haveI := rat.discrete_linear_ordered_field; apply_instance⟩ :=
begin
  by_cases int_bdd : ∃ B, ∀ z : ℕ, abv z ≤ B,
  {
    apply or.inr,
    apply' rat_abs_val_bdd_padic,
  },
  {
    apply or.inl,
    exact rat_abs_val_unbdd_real _ _ _ _,
  }
end
